// Copyright (c) 2012, Ladislav Thon. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.

part of deequery;

class _EagerQueryImpl<E> implements Query<E> {
  List<E> _list;

  _EagerQueryImpl(Iterable<E> iterable) : _list = new List.from(iterable);

  Iterator<E> get iterator => _list.iterator;

  Query<E> append(Iterable<E> other) {
    _list.addAll(new List<E>.from(other));
    return this;
  }

  // I hate it, but it's O(N**2) now...
  Query<E> unique() {
    bool contains(List<E> list, E element) => list.any((e) => e == element);

    var result = <E>[];
    for (var e in _list) {
      if (!contains(result, e)) {
        result.add(e);
      }
    }

    _list = result;
    return this;
  }

  Query<E> filter(predicate) {
    _list = _list.where(positivePredicate(predicate));
    return this;
  }

  Query<E> reject(predicate) {
    _list = _list.where(negativePredicate(predicate));
    return this;
  }

  Query<E> sort([int compare(E a, E b)]) {
    if (compare == null) {
      compare = (E a, E b) => a.compareTo(b);
    }

    _list.sort(compare);
    return this;
  }

  Query<E> sortBy(extractCompareKey(E e)) {
    Comparable comparableFromKey(key) {
      if (key is Comparable) {
        return key;
      } else if (key is List<Comparable>) {
        return new _CompoundComparable(key);
      } else {
        throw new ArgumentError("Compare key must be either Comparable or List<Comparable>");
      }
    }

    _list.sort((E a, E b) {
      var aCmp = comparableFromKey(extractCompareKey(a));
      var bCmp = comparableFromKey(extractCompareKey(b));
      return aCmp.compareTo(bCmp);
    });
    return this;
  }

  Query<E> reverse() {
    var result = <E>[];
    for (var i = _list.length - 1; i >= 0; i--) {
      result.add(_list[i]);
    }

    _list = result;
    return this;
  }

  Query/*<F>*/ map(/*F*/ transform(E e)) {
    var result = /*<F>*/[];
    for (var e in _list) {
      result.add(transform(e));
    }

    return new _EagerQueryImpl/*<F>*/(result);
  }

  Query/*<F>*/ mapToMany(Iterable/*<F>*/ expand(E e)) {
    var result = /*<F>*/[];
    for (var e in _list) {
      result.addAll(expand(e));
    }

    return new _EagerQueryImpl/*<F>*/(result);
  }

  Query<E> skip(int count) {
    if (count >= _list.length) {
      _list = <E>[];
    } else {
      _list = _list.sublist(count);
    }
    return this;
  }

  Query<E> take(int count) {
    if (count > _list.length) {
      count = _list.length;
    }

    _list = _list.sublist(0, count);
    return this;
  }

  int _countFirst(predicate) {
    var p = positivePredicate(predicate);

    var count = 0;
    for (var e in _list) {
      if (p(e)) {
        count += 1;
      } else {
        break;
      }
    }
    return count;
  }

  Query<E> skipWhile(predicate) => skip(_countFirst(predicate));

  Query<E> takeWhile(predicate) => take(_countFirst(predicate));

  bool get isEmpty => _list.isEmpty;

  bool any(predicate) => _list.any(positivePredicate(predicate));

  bool every(predicate) => _list.every(positivePredicate(predicate));

  bool contains(E e) => _list.contains(e);

  E findFirst(predicate) {
    var p = positivePredicate(predicate);
    for (var e in _list) {
      if (p(e)) {
        return e;
      }
    }

    return null; // TODO what else?
  }

  /*F*/ fold(/*F*/ seed, /*F*/ aggregate(/*F*/ accumulator, E e)) {
    var acc = seed;
    for (var e in _list) {
      acc = aggregate(acc, e);
    }

    return acc;
  }

  /*F*/ foldRight(/*F*/ seed, /*F*/ aggregate(/*F*/ accumulator, E e)) {
    var acc = seed;
    for (var i = _list.length - 1; i >= 0; i--) {
      acc = aggregate(acc, _list[i]);
    }

    return acc;
  }

  E reduce(E aggregate(E accumulator, E e)) {
    if (_list.isEmpty) {
      return null;
    }

    var acc = _list[0];
    for (var i = 1; i < _list.length; i++) {
      acc = aggregate(acc, _list[i]);
    }
    return acc;
  }

  E reduceRight(E aggregate(E accumulator, E e)) {
    if (_list.isEmpty) {
      return null;
    }

    var acc = _list[_list.length - 1];
    for (var i = _list.length - 2; i >= 0; i--) {
      acc = aggregate(acc, _list[i]);
    }
    return acc;
  }

  E max() {
    if (_list.isEmpty) {
      return null;
    }

    var currentMax = _list[0];
    for (var e in _list) {
      if (e > currentMax) {
        currentMax = e;
      }
    }

    return currentMax;
  }

  E min() {
    if (_list.isEmpty) {
      return null;
    }

    var currentMin = _list[0];
    for (var e in _list) {
      if (e < currentMin) {
        currentMin = e;
      }
    }

    return currentMin;
  }

  E sum() {
    if (_list.isEmpty) {
      return null;
    }

    var acc = _list[0];
    if (_list.length > 1) {
      for (var i = 1; i < _list.length; i++) {
        acc += _list[i];
      }
    }

    return acc;
  }

  String asString(String separator, {bool skipNulls: true}) {
    var strings = new List<String>();
    for (var o in _list) {
      if (skipNulls && o == null) {
        continue;
      }
      strings.add(o.toString());
    }

    return strings.join(separator);
  }

  E operator [](int index) => index >= 0 ? _list[index] : _list[_list.length + index];
}
