// Copyright (c) 2012, Ladislav Thon. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.

part of deequery;

class _LazyQueryImpl<E> implements Query<E> {
  final Iterable<E> _iterable;

  _LazyQueryImpl(this._iterable);

  Iterator<E> get iterator => _iterable.iterator;

  Query<E> get immediate => new _EagerQueryImpl<E>(_iterable);

  Query<E> append(Iterable<E> other) {
    return new _LazyQueryImpl<E>(new _CompoundIterable(_iterable, other));
  }

  Query<E> unique() {
    return new _LazyQueryImpl<E>(new _UniqueIterable(_iterable));
  }

  Query<E> filter(predicate) {
    return new _LazyQueryImpl<E>(new _FilteredIterable(_iterable, positivePredicate(predicate)));
  }

  Query<E> reject(predicate) {
    return new _LazyQueryImpl<E>(new _FilteredIterable(_iterable, negativePredicate(predicate)));
  }

  Query<E> sort([int compare(E a, E b) = null]) => immediate.sort(compare);

  Query<E> sortBy(extractCompareKey(E e)) => immediate.sortBy(extractCompareKey);

  Query<E> reverse() => immediate.reverse();

  Query/*<F>*/ map(/*F*/ transform(E e)) {
    return new _LazyQueryImpl/*<F>*/(new _MapIterable(_iterable, transform));
  }

  Query/*<F>*/ mapToMany(Iterable/*<F>*/ expand(E e)) {
    return new _LazyQueryImpl/*<F>*/(new _MapToManyIterable(_iterable, expand));
  }

  Query<E> skip(int count) {
    return new _LazyQueryImpl<E>(new _SkipIterable(_iterable, count));
  }

  Query<E> take(int count) {
    return new _LazyQueryImpl<E>(new _TakeIterable(_iterable, count));
  }

  Query<E> skipWhile(predicate) {
    return new _LazyQueryImpl<E>(new _SkipWhileIterable(_iterable, positivePredicate(predicate)));
  }

  Query<E> takeWhile(predicate) {
    return new _LazyQueryImpl<E>(new _TakeWhileIterable(_iterable, positivePredicate(predicate)));
  }

  bool get isEmpty => immediate.isEmpty;

  bool any(predicate) {
    var p = positivePredicate(predicate);
    for (var e in _iterable) {
      if (p(e)) {
        return true;
      }
    }
    return false;
  }

  bool every(predicate) => immediate.every(predicate);

  bool contains(E e) => any((item) => e == item);

  E findFirst(predicate) {
    final p = positivePredicate(predicate);
    for (var e in _iterable) {
      if (p(e)) {
        return e;
      }
    }

    return null; // TODO what else?
  }

  /*F*/ fold(/*F*/ seed, /*F*/ aggregate(/*F*/ Object accumulator, E e)) => immediate.fold(seed, aggregate);

  /*F*/ foldRight(/*F*/ seed, /*F*/ aggregate(/*F*/ Object accumulator, E e)) => immediate.foldRight(seed, aggregate);

  E reduce(E aggregate(E accumulator, E e)) => immediate.reduce(aggregate);

  E reduceRight(E aggregate(E accumulator, E e)) => immediate.reduceRight(aggregate);

  E max() => immediate.max();

  E min() => immediate.min();

  E sum() => immediate.sum();

  String asString(String separator, {bool skipNulls: true}) => immediate.asString(separator, skipNulls: skipNulls);

  E operator [](int index) {
    if (index >= 0) {
      var it = this.iterator;
      while (index > 0 && it.moveNext()) {
        index--;
      }
      if (index > 0 || !it.moveNext()) {
        throw new RangeError(index);
      }
      return it.current;
    } else {
      return immediate[index];
    }
  }
}

// append

class _CompoundIterator<E> implements Iterator<E> {
  Iterator<E> first, second;
  bool inFirst = true;

  _CompoundIterator(this.first, this.second);

  bool moveNext() {
    if (inFirst) {
      var result = first.moveNext();
      if (!result) {
        inFirst = false;
        result = second.moveNext();
      }
      return result;
    } else {
      return second.moveNext();
    }
  }

  E get current => inFirst ? first.current : second.current;
}

class _CompoundIterable<E> implements Iterable<E> {
  Iterable<E> first, second;

  _CompoundIterable(this.first, this.second);

  Iterator<E> get iterator => new _CompoundIterator(first.iterator, second.iterator);
}

// unique

// I hate it, but it's O(N**2) now...
class _UniqueIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  List<E> alreadySeen = <E>[];

  E currentValue;

  _UniqueIterator(this.iterator);

  bool moveNext() {
    while (iterator.moveNext()) {
      E value = iterator.current;
      if (!alreadySeen.contains(value)) {
        currentValue = value;
        alreadySeen.add(value);
        return true;
      }
    }

    currentValue = null;
    return false;
  }

  E get current => currentValue;
}

class _UniqueIterable<E> implements Iterable<E> {
  Iterable<E> iterable;

  _UniqueIterable(this.iterable);

  Iterator<E> get iterator => new _UniqueIterator(iterable.iterator);
}

// filter, reject

class _FilteredIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  Predicate predicate;

  _FilteredIterator(this.iterator, this.predicate);

  bool moveNext() {
    while (iterator.moveNext()) {
      if (predicate(iterator.current)) {
        return true;
      }
    }

    return false;
  }

  E get current => iterator.current;
}

class _FilteredIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  Predicate predicate;

  _FilteredIterable(this.iterable, this.predicate);

  Iterator<E> get iterator => new _FilteredIterator(iterable.iterator, predicate);
}

// map

class _MapIterator<E, F> implements Iterator<F> {
  Iterator<E> iterator;
  var transform;

  _MapIterator(this.iterator, F this.transform(E e));

  bool moveNext() => iterator.moveNext();

  F get current => transform(iterator.current);
}

class _MapIterable<E, F> implements Iterable<F> {
  Iterable<E> iterable;
  var transform;

  _MapIterable(this.iterable, F this.transform(E e));

  Iterator<F> get iterator => new _MapIterator<E, F>(iterable.iterator, transform);
}

// mapToMany

class _MapToManyIterator<E, F> implements Iterator<F> {
  Iterator<E> iterator;
  var expand;
  Iterator<F> currentIterator;
  bool started = false;

  _MapToManyIterator(this.iterator, Iterable<F> this.expand(E e));

  bool moveNext() {
    while (!started || !currentIterator.moveNext()) {
      started = true;
      currentIterator = null;
      if (iterator.moveNext()) {
        currentIterator = expand(iterator.current).iterator;
      } else {
        return false;
      }
    }
    return true;
  }

  F get current => currentIterator == null ? null : currentIterator.current;
}

class _MapToManyIterable<E, F> implements Iterable<F> {
  Iterable<E> iterable;
  var expand;

  _MapToManyIterable(this.iterable, Iterable<F> this.expand(E e));

  Iterator<F> get iterator => new _MapToManyIterator<E, F>(iterable.iterator, expand);
}

// skip

class _SkipIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  int toSkip;

  _SkipIterator(this.iterator, this.toSkip);

  bool moveNext() {
    if (toSkip > 0) {
      for (var i = 0; i < toSkip; i++) {
        iterator.moveNext();
      }
      toSkip = 0;
    }

    return iterator.moveNext();
  }

  E get current => iterator.current;
}

class _SkipIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  int count;

  _SkipIterable(this.iterable, this.count);

  Iterator<E> get iterator => new _SkipIterator(iterable.iterator, count);
}

// take

class _TakeIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  int toTake;

  _TakeIterator(this.iterator, this.toTake);

  bool moveNext() {
    if (toTake > 0 && iterator.moveNext()) {
      toTake--;
      return true;
    }

    toTake = -1;
    return false;
  }

  E get current => toTake < 0 ? null : iterator.current;
}

class _TakeIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  int count;

  _TakeIterable(this.iterable, this.count);

  Iterator<E> get iterator => new _TakeIterator(iterable.iterator, count);
}

// skip while

class _SkipWhileIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  Predicate predicate;
  bool skipped = false;

  _SkipWhileIterator(this.iterator, this.predicate);

  bool moveNext() {
    if (!skipped) {
      while (iterator.moveNext()) {
        if (!predicate(iterator.current)) {
          skipped = true;
          return true;
        }
      }

      skipped = true;
      return false;
    }

    return iterator.moveNext();
  }

  E get current => iterator.current;
}

class _SkipWhileIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  Predicate predicate;

  _SkipWhileIterable(this.iterable, this.predicate);

  Iterator<E> get iterator => new _SkipWhileIterator(iterable.iterator, predicate);
}

// take while

class _TakeWhileIterator<E> implements Iterator<E> {
  Iterator<E> iterator;
  Predicate predicate;
  bool taken;

  _TakeWhileIterator(this.iterator, this.predicate) : taken = false;

  bool moveNext() {
    if (taken) {
      return false;
    }

    if (iterator.moveNext() && predicate(iterator.current)) {
      return true;
    }

    taken = true;
    return false;
  }

  E get current => taken ? null : iterator.current;
}

class _TakeWhileIterable<E> implements Iterable<E> {
  Iterable<E> iterable;
  Predicate predicate;

  _TakeWhileIterable(this.iterable, this.predicate);

  Iterator<E> get iterator => new _TakeWhileIterator(iterable.iterator, predicate);
}
